from functools                   import wraps
from re                          import sub
from sys                         import stderr
from prompt_toolkit              import print_formatted_text, HTML, prompt
from prompt_toolkit.auto_suggest import AutoSuggest, Suggestion

import mekpie.debug as debug

config_layer = 0

class SuggestFromOptions(AutoSuggest):

    def __init__(self, options):
        self.options = options

    def get_suggestion(self, buffer, document):
        text = document.text.rsplit('\n', 1)[-1]
        if text.strip():
            for option in self.options:
                if option.startswith(text):
                    return Suggestion(option[len(text):])

def cli_config(name):
    def cli_config_decorator(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            global config_layer
            tell(f'┌ Configuring {name}...')
            config_layer += 1
            result = fn(*args, **kwargs)
            config_layer -= 1
            tell(f'└ {name} configured!')
            return result
        return wrapper
    return cli_config_decorator

def fix_text(text):
    # replace keywordss
    text = sub(r'`(.*?)`', r'<ansimagenta>\g<1></ansimagenta>', text)
    # replace errors
    text = sub(r'!! (.*)', r'<ansired>\g<1></ansired>', text)
    return text

def tell(message, out=True, **kwargs):
    header  = '│ ' * config_layer
    message = message.strip().replace('\n', f'\n{header}')
    result  = HTML(fix_text(f'{header}{message} '))
    if (out):
        print_formatted_text(result, **kwargs)
    return result

def yell(message, out=True, **kwargs):
    header  = '│ ' * config_layer
    message = message.strip().replace('\n', f'\n{header}')
    result  = HTML(fix_text(f'{header}<ansired>{message}</ansired>'))
    if (out):
        print_formatted_text(result, **kwargs)
    return result

def panic(message=None, **kwargs):
    if message is None:
        exit(1)
    tell(f'{message}', file=stderr, end='\n\n', **kwargs)

    if debug.debug:
        raise Exception('Debug Exception')
    else:
        exit(1)

def log(value):
    message = (str(value).strip()
        .replace('>', '')
        .replace('<', '')
    )
    if debug.debug:
        tell(f'<ansiblue>[LOG] {message}</ansiblue>')
    return value

def ask(question, default=None, options=None, validator=None):
    message, error = question
    suggestor = SuggestFromOptions(options) if options else None
    while True:
        footer = ':' if default is None else f' (default `{default}`):'
        value = prompt(tell(f'{message}{footer}', False), auto_suggest=suggestor)
        if value == '' and default:
            value = default
        if ((options and value in options)
            or (validator and validator(value))
            or (not validator and not options)
        ):
            tell(f'Selected `{value}`.')
            return value
        yell(error.format(value))


from .definitions import CompilerConfig, CompilerConfig
from .runner      import autodetect
from .cli         import cli_config, ask, tell

def emscripten():

    def cc_once(cfg):
        lib_flags     = ['-l' + lib for lib in cfg.libs]
        include_flags = ['-I' + inc for inc in cfg.includes]
        return cfg.flags + include_flags + lib_flags
    
    def cc_compile(cfg, source):
        return source

    def cc_link(cfg, main, objects):
        path = cfg.targetpath(main) + '.js'
        cfg.run([
            'emcc',
            main,
            *objects,
            *cfg.once,
            '-o', path,
        ])
        return path

    def cc_run(cfg, exe):
        cfg.run(['node', exe, *cfg.options.programargs])

    def cc_debug(cfg, exe):
        cfg.run(['node', 'inspect', exe])

    return CompilerConfig(
        name         = 'emscripten',
        compile      = cc_compile,
        link         = cc_link,
        run          = cc_run,
        debug        = cc_debug,
        once         = cc_once,
        ccsource     = f'emscripten()',
        debugflags   = ['-Wall', '-g'],
        releaseflags = ['-Wall', '-O'],
    )

@cli_config('emscripten')
def config_emscripten():
    return emscripten()


required = { 'require_me': True }

class RecordClass(dict):

    def __init__(self, defaults, **kwargs):
        self.update(defaults)
        init_values = dict(**kwargs)
        for key, value in init_values.items():
            if key in self:
                self[key] = value
            else:
                raise KeyError(key)
        for key, value in self.items():
            if value == required:
                raise KeyError(f'{key} is required!')

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        if key not in self:
            raise KeyError(key)
        self[key] = value

class Record:

    def __init__(self, defaults):
        self.defaults = defaults
        self.keys     = defaults.keys()

    def __call__(self, **kwargs):
        return RecordClass(self.defaults, **kwargs)



from os.path import isdir, curdir, join
from os      import chdir
from os.path import basename, abspath, exists

import mekpie.messages as messages

from .cli           import panic
from .util          import empty, first, smkdir, exec_str
from .config        import config_from_str
from .cli           import cli_config, ask, tell
from .definitions   import Config, DEFAULT_MEKPY
from .cc_gcc_clang  import config_gcc_clang
from .cc_avr_gcc    import config_avr_gcc
from .cc_emscripten import config_emscripten
from .structure     import (
    set_project_path,
    get_project_path,
    get_mekpy_path,
    get_src_path,
    get_main_path,
    get_test_path,
    get_includes_path,
    get_target_path,
    get_target_debug_path,
    get_target_release_path,
)

@cli_config('mekpie')
def config_mekpie(options):
    name = ask(messages.mekpie_config_name,
        default   = first(options.commandargs),
        validator = lambda s : not empty(s),
    )
    tell(messages.compiler_configs)
    cc = {
        'gcc_clang'  : config_gcc_clang,
        'avr_gcc'    : config_avr_gcc,
        'emscripten' : config_emscripten,
    }[ask(messages.mekpie_config_cc,
        default = 'gcc_clang',
        options = ['gcc_clang', 'avr_gcc', 'emscripten'],
    )]()
    return name, cc

def command_new(options):
    name, cc = config_mekpie(options)
    check_name(name)
    create_project_directory(name)
    create_mekpy(name, cc)
    create_src(name, cc)
    create_tests()
    create_includes()
    create_target()
    config_from_str(options, get_mekpy_source(name, cc))
    tell(messages.created.format(name).strip())

def command_init(options):
    if not first(options.commandargs):
        options.commandargs.append(basename(abspath(curdir)))
    name, cc = config_mekpie(options)
    check_name(name)
    create_mekpy(name, cc)
    create_src(name, cc)
    create_tests()
    create_includes()
    create_target()
    config_from_str(options, get_mekpy_source(name, cc))
    tell(messages.initialized.format(name).strip())

def check_name(name):
    if not name:
        panic(messages.name_cannot_be_empty)
    if isdir(name):
        panic(messages.name_cannot_already_exist.format(name))

def create_project_directory(name):
    set_project_path(name)
    smkdir(get_project_path())

def create_mekpy(name, cc):
    if not exists(get_mekpy_path()):
        with open(get_mekpy_path(), 'w+') as rsc:
            rsc.write(get_mekpy_source(name, cc))

def create_src(name, cc):
    smkdir(get_src_path())
    if not exists(get_main_path(name + '.c')):
        with open(get_main_path(name + '.c'), 'w+') as rsc:
            rsc.write(cc.csource)

def create_tests():
    smkdir(get_test_path())

def create_includes():
    smkdir(get_includes_path())

def create_target():
    smkdir(get_target_path())
    smkdir(get_target_release_path())
    smkdir(get_target_debug_path())

def get_mekpy_source(name, cc):
    return DEFAULT_MEKPY.format(
        name, 
        name + '.c', 
        str(cc.ccsource),
        str(cc.releaseflags),
        str(cc.debugflags),
    )


from .record import Record, required

DEFAULT_MEKPY = '''
# This is a standard configuration file for mekpie

# the name of your project
name = '{}' 
# the .c file containing `main`
main = '{}'
# the c compiler configuration to use (gcc_clang, avr_gcc, or emscripten)
cc = {}
# any libraries to load
libs = []
# additional compiler flags
flags = []

if options.release:
    flags += {}
else:
    flags += {}
'''

MAIN = '''
#include <stdio.h>
#include <stdlib.h>

int main() {
    puts("Hello, World!");
    return EXIT_SUCCESS;
}
'''.strip()

VERSION = '0.1.2'

Option = Record({
    'names'   : required, 
    'nargs'   : None,
    'handler' : None,
})

Options = Record({
    'quiet'       : False,     # -q --quiet
    'release'     : False,     # -r --release
    'developer'   : False,     # -d --developer
    'mode'        : False,     # -m --mode
    'changedir'   : False,     # -c --changedir
    'command'     : None,      # new, init, clean, build, run, test, debug
    'commandargs' : [],        # <optionargs...>
    'programargs' : [],        # -- <programargs...>
})

Config = Record({
    'name'       : None, # Project name
    'main'       : None, # Entry point
    'libs'       : [],   # Libraries to link
    'cc'         : None, # Compiler Configuration
    'flags'      : [],   # User compiler flags
    'options'    : None, # The provided command line options
    'includes'   : [],   # Include directories
    'once'       : None, # Provided by plugin
    'targetpath' : None, # Function that providees target path
    'run'        : None, # Provides a run function with correct flags
})

CompilerConfig = Record({
    'name'         : required, # Name of the config
    'compile'      : required, # Compilation function
    'link'         : required, # Linking function
    'run'          : required, # Runnning function
    'debug'        : required, # Debug function
    'once'         : (lambda *args : None),
    'ccsource'     : None,     # Mekpy source for project creation
    'csource'      : MAIN,     # csource, by default hello world
    'debugflags'   : [],       # Default debug flags    
    'releaseflags' : [],       # Default release flags
})